# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent


class BaseFileHandler(BaseSubcomponent):
    
    supported_extensions = 'csv', 'xlsx', 'png', 'jpg', 'jpeg'
    
    def __init__(self, data_viewer):
        
        self._data_viewer = data_viewer
        self.setup(data_viewer)

    def open_in_workspace(self, code, path, tmpl):
        
        symbol = self._unique_symbol(tmpl)
        self.extension_manager.fire(
            'jupyter_run_code',
            code=code.format(path=path, symbol=symbol)
        )
        return symbol

    def _unique_symbol(self, tmpl):
        
        workspace_list = self.extension_manager.provide(
            'jupyter_list_workspace_globals'
        )
        i = 0
        while tmpl.format(i) in workspace_list:
            i += 1
        return tmpl.format(i)



from jupyter_notebook_cell_parsers.parse_nbformat import (
    cells_to_notebook,
    notebook_to_code
)
from jupyter_notebook_cell_parsers.parse_python import parse_python

NO_CELLS = '''print(1)
print(2)
'''
NO_CELLS_OUTPUT = []
ONE_SIMPLE_COMMENT = '''
"""
Comment 3
Comment 4
"""

'''
ONE_SIMPLE_COMMENT_OUTPUT = [{
    'cell_type': 'markdown',
    'source': 'Comment 3\nComment 4',
    'start': 1,
    'end': 29
}]
TWO_SIMPLE_CELLS = '''"""
Comment 5
"""
print('Code 6')
"""
Comment 7
"""
print('Code 8')
'''
TWO_SIMPLE_CELLS_OUTPUT = [
    {
        'cell_type': 'markdown',
        'source': 'Comment 5',
        'start': 0,
        'end': 18
    },
    {
        'cell_type': 'code',
        'source': "print('Code 6')",
        'start': 18,
        'end': 34
    },
    {
        'cell_type': 'markdown',
        'source': 'Comment 7',
        'start': 34,
        'end': 52
    },
    {
        'cell_type': 'code',
        'source': "print('Code 8')",
        'start': 52,
        'end': 68
    }
]
SIMPLE_CELL_WITH_OUTPUT = '''"""
Comment 9
"""

print('Code A')

# % output
# Code A
print('Code B')

# % output
# Code B
# ![](figure.png)
'''
SIMPLE_CELL_WITH_OUTPUT_OUTPUT = [
    {
        'cell_type': 'markdown',
        'source': 'Comment 9',
        'start': 0,
        'end': 18
    },
    {
        'cell_type': 'code',
        'source': "print('Code A')",
        'execution_count': 1,
        'start': 18,
        'end': 35,
        'output_type': 'execute_result',
        'outputs': '# Code A'
    },
    {
        'cell_type': 'code',
        'source': "print('Code B')",
        'execution_count': 2,
        'start': 56,
        'end': 72,
        'output_type': 'execute_result',
        'outputs': '# Code B\n# ![](figure.png)'
    }
]
BASIC_SPYDER_CELL = '''# %%
print(1)
print(2)
'''
BASIC_SPYDER_CELL_OUTPUT = [
    {
        'cell_type': 'code',
        'source': "print(1)\nprint(2)",
        'start': 5,
        'end': 22
    }
]

SPYDER_CELL_WITH_OUTPUT = '''# %%
# Comment 10

print('Code C')

# % output
# Code C
print('Code D')

# % output
# Code D
# ![](figure.png)
'''
SPYDER_CELL_WITH_OUTPUT_OUTPUT = [
    {
        'cell_type': 'markdown',
        'source': 'Comment 10',
        'start': 5,
        'end': 17
    },
    {
        'cell_type': 'code',
        'source': "print('Code C')",
        'execution_count': 1,
        'start': 18,
        'end': 35,
        'output_type': 'execute_result',
        'outputs': '# Code C'
    },
    {
        'cell_type': 'code',
        'source': "print('Code D')",
        'execution_count': 2,
        'start': 56,
        'end': 72,
        'output_type': 'execute_result',
        'outputs': '# Code D\n# ![](figure.png)'
    }
]


def _check_parser(fnc, code, result=None):
    
    cells = fnc(code)
    for cell in cells:
        # Check whether the start and end positions are consistent with the
        # text. Code cells are stripped of whitespaces, and comment cells
        # are stripped of triple quotes.
        if cell['cell_type'] == 'code':
            assert(code[cell['start']:cell['end']].strip() == cell['source'])
        elif cell['cell_type'] == 'markdown':
            code_chunk = code[cell['start']:cell['end']]
            if not code_chunk.startswith('#'):
                assert code_chunk == '"""\n{}\n"""\n'.format(cell['source'])
    # Assert that the full code matches the expected output
    if result is not None:
        assert(cells == result)


def test_python_parser():
    
    _check_parser(
        parse_python,
        TWO_SIMPLE_CELLS,
        TWO_SIMPLE_CELLS_OUTPUT
    )
    _check_parser(
        parse_python,
        NO_CELLS,
        NO_CELLS_OUTPUT
    )
    _check_parser(
        parse_python,
        ONE_SIMPLE_COMMENT,
        ONE_SIMPLE_COMMENT_OUTPUT
    )
    _check_parser(
        parse_python,
        SIMPLE_CELL_WITH_OUTPUT,
        SIMPLE_CELL_WITH_OUTPUT_OUTPUT
    )
    _check_parser(
        parse_python,
        BASIC_SPYDER_CELL,
        BASIC_SPYDER_CELL_OUTPUT
    )
    _check_parser(
        parse_python,
        SPYDER_CELL_WITH_OUTPUT,
        SPYDER_CELL_WITH_OUTPUT_OUTPUT
    )


def test_nbformat_parser():

    cells = parse_python(SIMPLE_CELL_WITH_OUTPUT)
    cells_to_notebook(cells, path='tmp.ipynb', language='python')
    language, code = notebook_to_code(
        'tmp.ipynb',
        (lambda fmt, content: 'figure.png')
    )
    assert(code == SIMPLE_CELL_WITH_OUTPUT)


if __name__ == '__main__':
    test_python_parser()
    test_nbformat_parser()



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import zmq
import time
import pickle
from libqtopensesame.extensions import BaseExtension
from libopensesame.oslogging import oslogger
from qtpy.QtCore import QTimer


BASE_PORT = 5555
POLL_INTERVAL_SHORT = 500
POLL_INTERVAL_LONG = 2000


class PythonDebugger(BaseExtension):
    
    def event_startup(self):
        
        self._server_started = False

    def _start_server(self):
        
        if self._server_started:
            return
        self._server_started = True
        self._zmq_context = zmq.Context()
        self._zmq_socket = self._zmq_context.socket(zmq.REP)
        self._zmq_port = BASE_PORT
        self._poll_interval = POLL_INTERVAL_SHORT
        # Loop to get the first available port
        while True:
            try:
                self._zmq_socket.bind('tcp://*:{}'.format(self._zmq_port))
            except zmq.error.ZMQError:
                self._zmq_port += 1
            else:
                break
        QTimer.singleShot(self._poll_interval, self._listen_for_pdb)
        
    def _listen_for_pdb(self):
         
        #  Wait for next request from client
        try:
            message = self._zmq_socket.recv_pyobj(flags=zmq.NOBLOCK)
        except zmq.error.ZMQError:
            QTimer.singleShot(self._poll_interval, self._listen_for_pdb)
            return
        try:
            self._zmq_socket.send(b'ok')
        except zmq.error.ZMQError:
            oslogger.error('failed to send reply')
        if not isinstance(message, dict) or 'type' not in message:
            oslogger.error('invalid debugger message: {}'.format(str(message)))
        elif message['type'] == 'pdb_start':
            oslogger.debug('debugger session started')
            self._poll_interval = POLL_INTERVAL_SHORT
        elif message['type'] == 'pdb_stop':
            oslogger.debug('debugger session stopped')
            self._poll_interval = POLL_INTERVAL_LONG
        elif message['type'] == 'pdb_frame':
            oslogger.debug('debugger frame received')
            self._poll_interval = POLL_INTERVAL_SHORT
            if os.path.exists(message['path']):
                self.extension_manager.fire(
                    'ide_open_file',
                    path=message['path'],
                    line_number=message['line']
                )
                try:
                    # Refocus the console, not the editor
                    self.extension_manager['JupyterConsole']._focus()
                except Exception:
                    pass
            self.extension_manager.fire(
                'workspace_update',
                name=self.extension_manager.provide('jupyter_workspace_name'),
                workspace_func=(lambda: message['workspace'])
            )
        else:
            oslogger.error(
                'invalid debugger message type: {}'.format(message['type'])
            )
        QTimer.singleShot(self._poll_interval, self._listen_for_pdb)

    def provide_python_debugger_code(self, path, breakpoints):
        
        self._start_server()
        with open(
            os.path.join(os.path.dirname(__file__), u'rapunzel_pdb.py')
        ) as f:
            silent_code = f.read()
        code = 'rpdb({}, port={}, breakpoints={})'.format(
            repr(path),
            self._zmq_port,
            breakpoints
        )
        return silent_code, code



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import importlib
import time
from qtpy.QtWidgets import QFileDialog, QMenu, QApplication
from libopensesame.oslogging import oslogger
from libopensesame import misc
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'JupyterNotebook', category=u'extension')


class JupyterNotebook(BaseExtension):
    """Handles various things related to JupyterLab:
    
    - Parses source code into cells, which are used for code execution by the
      OpenSesameIDE extension.
    - Allows code to be exported to, and imported from, ipynb format.
    - Allows JupyterLab to launched.
    """
    
    preferences_ui = 'extensions.JupyterNotebook.preferences'

    def event_startup(self):
        self._widget = None
        
    def event_ide_menubar_initialized(self, menubar):
        self._menu = QMenu(_('Import and export'), self.main_window)
        self._menu.setIcon(self.theme.qicon('text-x-script'))
        self._menu.addAction(self.qaction(
            u'document-open', _('Import notebook'), self._import_ipynb,
        ))
        self._menu.addSeparator()
        self._menu.addAction(self.qaction(
            u'document-save', _('Export notebook'), self._export_ipynb
        ))
        self._menu.addAction(self.qaction(
            u'document-save', _('Export pdf'), self._export_pdf
        ))
        self._menu.addAction(self.qaction(
            u'document-save', _('Export html'), self._export_html
        ))
        self._menu.addAction(self.qaction(
            u'document-save', _('Export docx'), self._export_docx
        ))
        menubar._menu_file.insertMenu(menubar._action_quit, self._menu)
        menubar._menu_file.insertSeparator(menubar._action_quit)

    def event_close(self):

        if self._widget is None:
            return
        self._widget.kill()

    def activate(self):

        self.tabwidget.add(self.widget(), self.icon(), self.label())

    def widget(self):
        """A simple widget for launching and killing jupyterlab."""
        
        if self._widget is None:
            self.set_busy()
            from jupyter_widget import LaunchJupyterLabWidget
            self._widget = LaunchJupyterLabWidget(self.main_window, self)
            self.set_busy(False)
        return self._widget

    def provide_jupyter_notebook_cells(self, code=u'', cell_types=None):
        """Dynamically loads a parser function for the language of the current
        editor, and uses this to parse the code into cells. The `cell_types`
        keyword allows specific kinds of cells to be returned.
        """
        
        language = self.extension_manager.provide('ide_current_language')
        try:
            m = importlib.import_module(
                'jupyter_notebook_cell_parsers.parse_{}'.format(language)
            )
        except ModuleNotFoundError:
            oslogger.debug('no cell parser for language {}'.format(language))
            return
        return getattr(m, 'parse_{}'.format(language))(code, cell_types)
    
    def provide_open_file_extension_ipynb(self):
        """A provider for directly opening .ipynb files."""
        
        return self._import_ipynb, _('Import as script')
    
    def _import_ipynb(self, path=None):
        """Import an ipynb file from path and return it as plain-text code."""
        
        from jupyter_notebook_cell_parsers import parse_nbformat
        
        if not isinstance(path, basestring):
            path = QFileDialog.getOpenFileName(
                self.main_window,
                _(u'Open Jupyter/ IPython Notebook'),
                filter=u'Notebooks (*.ipynb)',
                directory=cfg.file_dialog_path
            )
        if isinstance(path, tuple):
            path = path[0]
        if not path:
            return
        cfg.file_dialog_path = os.path.dirname(path)
        try:
            language, code = parse_nbformat.notebook_to_code(
                path,
                self.extension_manager.provide('image_writer')
            )
        except Exception as e:
            self.extension_manager.fire(
                u'notify',
                message=_(u'Failed to read notebook. See console for details.')
            )
            self.console.write(e)
            return
        if language.lower() == 'r':
            ext = '.R'
        elif language.lower() == 'python':
            ext = '.py'
        else:
            ext = None
        self.extension_manager.fire(u'ide_new_file', source=code, ext=ext)
        self.extension_manager.fire(u'image_annotations_detect', code=code)

    def _export(self, dialog_title, dialog_filter, ext):
        """Export the code to an .ipynb file."""
        
        from jupyter_notebook_cell_parsers import parse_nbformat

        language = self.extension_manager.provide('ide_current_language')
        path = QFileDialog.getSaveFileName(
            self.main_window,
            dialog_title,
            filter=dialog_filter,
            directory=cfg.file_dialog_path
        )
        if isinstance(path, tuple):
            path = path[0]
        if not path:
            return None, None
        cfg.file_dialog_path = os.path.dirname(path)
        if not path.lower().endswith(ext):
            path += ext
        if path.lower().endswith('.ipynb'):
            ipynb_path = path
        else:
            ipynb_path = os.path.splitext(path)[0] + '.ipynb'
        parse_nbformat.cells_to_notebook(
            self.provide_jupyter_notebook_cells(
                self.extension_manager.provide('ide_current_source')
            ),
            ipynb_path,
            language
        )
        return path, ipynb_path
    
    def _run(self, cmd):
        
        self.main_window.set_busy(True)
        self.extension_manager.fire('jupyter_run_system_command', cmd=cmd)
        while self.extension_manager.provide('jupyter_kernel_running'):
            QApplication.processEvents()
            time.sleep(0.1)
        self.main_window.set_busy(False)
        
    def _to_html(self, ipynb_path):
        
        self._run('jupyter nbconvert "{}" --to html'.format(ipynb_path))
        return os.path.splitext(ipynb_path)[0] + '.html'
        
    def _export_ipynb(self):
        """Export the code to an .ipynb file."""
        
        self._export(
            dialog_title=_(u'Export to Jupyter/ IPython Notebook'),
            dialog_filter=u'Notebooks (*.ipynb)',
            ext='.ipynb'
        )

    def _export_pdf(self):
        path, ipynb_path = self._export(
            dialog_title=_(u'Export to pdf'),
            dialog_filter=u'pdf (*.pdf)',
            ext='.pdf'
        )
        if path is None:
            return
        html_path = self._to_html(ipynb_path)
        self._run('pandoc "{}" -o "{}"'.format(html_path, path))
        misc.open_url(path)
        
    def _export_html(self):
        path, ipynb_path = self._export(
            dialog_title=_(u'Export to HTML'),
            dialog_filter=u'html (*.html)',
            ext='.html'
        )
        if path is None:
            return
        html_path = self._to_html(ipynb_path)
        misc.open_url(html_path)

    def _export_docx(self):
        path, ipynb_path = self._export(
            dialog_title=_(u'Export to .docx'),
            dialog_filter=u'docx (*.docx)',
            ext='.docx'
        )
        if path is None:
            return
        html_path = self._to_html(ipynb_path)
        self._run('pandoc "{}" -o "{}"'.format(html_path, path))
        misc.open_url(path)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import re
import os
import uuid
import yaml
from qtpy import QtWidgets, QtCore
from pyqode.core import modes
from libopensesame.oslogging import oslogger
from libopensesame import misc
from libqtopensesame.misc.config import cfg
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'ImageAnnotations', category=u'extension')

OUTPUT_MARKER = '# % output'
HOURGLASSES = ['# ○\n', '# ◔\n', '# ◑\n', '# ◕\n', '# ●\n']
IMG_PATTERN = r'!\[\]\((?P<path>.+)\)'


class ImageAnnotations(BaseExtension):
    """Receives images that result from code execution (sent by JupyterConsole)
    and shows these as annotations next to the executed code. Relies on
    ImageAnnotationsMode and ImageAnnotationsPanel from pyqode.core.
    """
    
    preferences_ui = 'extensions.ImageAnnotations.preferences'
    re_strip_ansi_codes = re.compile(chr(27) + r'\[[0-9;]+m')
    
    def event_startup(self):
        """Creates a folder for the image files if it doesn't exist, and
        restores the image annotation info.
        """
        self._capturing = False
        self._hourglass = 0
        self._editor = None
        self.enable_capture = False
        self._image_folder = os.path.join(
            self.main_window.home_folder,
            '.opensesame',
            'image_annotations'
        )
        if not os.path.exists(self._image_folder):
            oslogger.debug('creating {}'.format(self._image_folder))
            os.mkdir(self._image_folder)
        try:
            self._image_annotations = safe_yaml_load(cfg.image_annotations)
            assert(isinstance(self._image_annotations, dict))
        except BaseException:
            oslogger.warning('failed to parse image annotations from config')
            self._image_annotations = {}
        # Remove annotations for non-existing images and convert single
        # annotations to lists of annotations so that there can be multiple
        # annotations for the same code. The `list()` wrapping is necessary
        # so that we can modify the objects while iterating through them.
        for file_path, file_annotations in list(
            self._image_annotations.items()
        ):
            if not isinstance(file_annotations, dict):
                self._image_annotations[file_path] = {}
                continue
            for code, paths in list(file_annotations.items()):
                if isinstance(paths, basestring):
                    paths = [paths]
                paths = [path for path in paths if os.path.exists(path)]
                if paths:
                    file_annotations[code] = paths
                else:
                    del file_annotations[code]
                    
    def event_setting_changed(self, setting, value):
        """Toggles capture of output when this changed in the menu."""
        if setting == 'image_annotations_enabled':
            if value:
                cfg.image_annotations_enabled = True
            else:
                cfg.image_annotations_enabled = False
                cfg.image_annotations_capture_output = False
        elif setting == 'image_annotations_capture_output':
            if value:
                cfg.image_annotations_enabled = True
                cfg.image_annotations_capture_output = True
            else:
                cfg.image_annotations_capture_output = False
        else:
            return
        editor = self.extension_manager.provide('ide_current_editor')
        if editor is None:
            return
        # If one of the settings is enabled, we add the mode to the editor. If
        # it is already installed, this will trigger a refresh.
        if cfg.image_annotations_enabled:
            self._add_mode(editor)
        # If the annotations are disabled altogether, then we remove the mode.
        else:
            self._remove_mode(editor)
        
    def event_register_editor(self, editor):
        """When an editor is registered, an ImageAnnotationsMode is added, but
        only if the editor already has an ImageAnnotationsPanel, which is
        managed by pyqode_manager.
        """
        if not cfg.image_annotations_enabled:
            return
        try:
            editor.panels.get('ImageAnnotationsPanel')
        except KeyError:
            return
        self._add_mode(editor)
        
    def _add_mode(self, editor):
        """Adds the image annotation mode to the editor if it wasn't already
        installed, and sets the annotations.
        """
        if 'ImageAnnotationsMode' in editor.modes.keys():
            mode = editor.modes.get('ImageAnnotationsMode')
        else:
            mode = modes.ImageAnnotationsMode()
            mode.annotation_clicked.connect(self._on_annotation_clicked)
            editor.modes.append(mode)
        self._set_annotations(editor, mode)
        
    def _remove_mode(self, editor):
        """Removes the image annotation mode from the editor."""
        if 'ImageAnnotationsMode' not in editor.modes.keys():
            return
        editor.modes.remove('ImageAnnotationsMode')

    def event_jupyter_run_code(self, code, editor=None):
        self._start_capture(editor)
        
    def provide_image_writer(self):
        return self._img_to_file
    
    def provide_image_annotations_capturing(self):
        if self._editor is None:
            return False
        return HOURGLASSES[self._hourglass] in self._editor.toPlainText()
    
    def event_image_annotations_detect(self, code):
        """Scans the code for markdown style images."""
        if not isinstance(code, basestring):
            return
        editor = self.extension_manager.provide('ide_current_editor')
        if editor is None:
            return
        try:
            mode = editor.modes.get('ImageAnnotationsMode')
        except KeyError:
            return
        for m in re.finditer(IMG_PATTERN, code):
            img_path = m.group('path')
            if editor.file.path not in self._image_annotations:
                self._image_annotations[editor.file.path] = {}
            img_code = '![]({})'.format(img_path)
            if img_code not in self._image_annotations[editor.file.path]:
                self._image_annotations[editor.file.path][img_code] = []
            self._image_annotations[editor.file.path][img_code].append(
                img_path
            )
        self._set_annotations(editor, mode)
        
    def event_image_annotations_clear_output(self):
        
        editor = self.extension_manager.provide('ide_current_editor')
        if editor is None:
            return
        code = editor.toPlainText()
        cells = self.extension_manager.provide(
            'jupyter_notebook_cells',
            code=code,
            cell_types=['code']
        )
        for cell in cells:
            if 'outputs' not in cell:
                continue
            # Replace the output and also clear up newlines before and after
            # the output. This avoids newlines from accumulating in the
            # document after repeated running and clearing the code.
            output = '{}\n{}'.format(OUTPUT_MARKER, cell['outputs'])
            while True:
                if '\n' + output in code:
                    output = '\n' + output
                elif output + '\n' in code:
                    output += '\n'
                else:
                    break
            code = code.replace(output, '\n\n')
        # We use a cursor rather than simply setting the text, so that this
        # action can be undone.
        cursor = editor.textCursor()
        cursor.select(cursor.Document)
        cursor.insertText(code)
        editor.setTextCursor(cursor)
        
    def event_ide_save_current_file_as(self, from_path, to_path):
        
        if from_path not in self._image_annotations:
            return
        self._image_annotations[to_path] = self._image_annotations[from_path]
        
    def _start_capture(self, editor):
        """Inserts a capture placeholder after the cursor when capturing is
        enabled.
        """
        if editor is None:
            self._editor = self.extension_manager.provide('ide_current_editor')
        else:
            self._editor = editor
        if self._editor is None:
            return
        if not cfg.image_annotations_capture_output:
            return
        if self._capturing:
            self.extension_manager.fire(
                'notify',
                message=_('Already capturing output')
            )
            return
        self._capturing = True
        self._has_captured = False
        self._code = None
        # The hourglass serves as a progress indicator, but also as a reference
        # to insert captured text before.
        self._hourglass = 0
        self._text_operation('\n' + HOURGLASSES[self._hourglass])
        QtCore.QTimer.singleShot(250, self._animate_hourglass)
        
    def _animate_hourglass(self):
        """Animates the hourglass."""
        if not self._capturing:
            self._text_operation(HOURGLASSES[self._hourglass], insert=False)
            return
        new_hourglass = (self._hourglass + 1) % len(HOURGLASSES)
        self._text_operation(
            HOURGLASSES[self._hourglass],
            insert=False,
            replace_by=HOURGLASSES[new_hourglass]
        )
        self._hourglass = new_hourglass
        QtCore.QTimer.singleShot(500, self._animate_hourglass)
    
    def event_jupyter_execute_finished(self):
        """Removes the capture placeholder when the Jupyter console is done
        executing.
        """
        self._capturing = False
        
    def _format_output(self, text):
        """Formats a block of output so that ansi color codes are removed and
        that each lines is quoted.
        """
        return '\n'.join([
            '# {}'.format(self.re_strip_ansi_codes.sub(' ', line))
            for line in text.splitlines()
        ])
        
    def event_jupyter_execute_result_text(self, text):
        """Gets text that is sent to the Jupyter console and captures it if
        capturing is enabled.
        """
        if not self._capturing:
            return
        if not self._has_captured:
            # Ignore empty lines at the start. Also ignore folder names, 
            # because changing the working directory annoyingly results in this
            # folder being printed out as the first result line.
            if not text.strip() or os.path.isdir(text.strip()):
                return
            self._text_operation(
                OUTPUT_MARKER,
                insert_position=self._editor.toPlainText().find(
                    HOURGLASSES[self._hourglass]
                )
            )
        self._has_captured = True
        formatted_output = self._format_output(text)
        if not formatted_output:
            return
        self._text_operation(
            formatted_output + '\n',
            insert_position=self._editor.toPlainText().find(
                HOURGLASSES[self._hourglass]
            )
        )
    
    def event_jupyter_execute_result_image(self, img, fmt, code):
        """Receives a new image annotation, where `img` is the image data,
        `fmt` is the format, and code is the code snippet that resulted in
        creating of the image.
        """
        if code is None:
            oslogger.warning('received None code')
            return
        try:
            mode = self._editor.modes.get('ImageAnnotationsMode')
        except (AttributeError, KeyError):
            return
        if self._editor.file.path not in self._image_annotations:
            self._image_annotations[self._editor.file.path] = {}
        if code not in self._image_annotations[self._editor.file.path]:
            self._image_annotations[self._editor.file.path][code] = []
        img_path = self._img_to_file(img, fmt)
        self._image_annotations[self._editor.file.path][code].append(img_path)
        if cfg.image_annotations_capture_output:
            md = '![]({})'.format(img_path)
            self._image_annotations[self._editor.file.path][md] = [img_path]
            self.event_jupyter_execute_result_text(md)
        self._set_annotations(self._editor, mode)
        
    def _set_annotations(self, editor, mode):
        """Sends the annotations to the AnnotationMode."""
        if cfg.image_annotations_capture_output:
            def annotation_filter(code): return code.startswith('![](')
        else:
            def annotation_filter(code): return not code.startswith('![](')
        cfg.image_annotations = yaml.dump(self._image_annotations)
        if editor.file.path in self._image_annotations:
            mode.set_annotations({
                editor.file.path: {
                    code: img_path
                    for (code, img_path)
                    in self._image_annotations[editor.file.path].items()
                    if code is not None and annotation_filter(code)
                }
            })
        else:
            mode.set_annotations({})
        
    def _img_to_file(self, img, fmt):
        """Save an image to file and return the filename. Inspired by Spyder's
        `figurebrowser.py`.
        """
        if fmt == 'image/svg+xml':
            ext = '.svg'
        elif fmt == 'image/png':
            ext = '.png'
        elif fmt == 'image/jpeg':
            ext = '.jpg'
        else:
            raise ValueError('invalid image format: {}'.format(fmt))
        path = os.path.join(self._image_folder, uuid.uuid4().hex + ext)
        oslogger.debug('writing image to {}'.format(path))
        if fmt == 'image/svg+xml' and isinstance(img, str):
            img = img.encode('utf-8')
        with open(path, 'wb') as fd:
            fd.write(img)
        return path
    
    def _open_annotation(self, path):
        """Opens a single annotation image."""
        misc.open_url(path)
        
    def _open_folder(self):
        """Opens the folder that contains the image annotations."""
        misc.open_url(self._image_folder)
        
    def _clear_annotation(self, path=None):
        """Clear a single image annotation, or all image annotations in `path`
        is `None`.
        """
        editor = self.extension_manager.provide('ide_current_editor')
        if editor is None:
            return
        try:
            mode = editor.modes.get('ImageAnnotationsMode')
        except KeyError:
            return
        if path is None:
            self._image_annotations[editor.file.path] = {}
        else:
            for code, paths in list(
                self._image_annotations[editor.file.path].items()
            ):
                if path in paths:
                    paths.remove(path)
                # If there are still annotations remaining, update the paths to
                # the annotations, otherwise delete the entry for the code
                # alltogether
                if paths:
                    self._image_annotations[editor.file.path][code] = paths
                else:
                    del self._image_annotations[editor.file.path][code]
                # If no annotations remain for the file, remove the entry for
                # the file alltogether.
                if not self._image_annotations[editor.file.path]:
                    del self._image_annotations[editor.file.path]
        self._set_annotations(editor, mode)

    def _on_annotation_clicked(self, msg, event):
        """Shows a context menu on the annotation."""
        if event.button() != QtCore.Qt.RightButton:
            return
        menu = QtWidgets.QMenu(self.main_window)
        menu.addAction(_('Open'), lambda: self._open_annotation(msg.path))
        menu.addSeparator()
        menu.addAction(_('Clear'), lambda: self._clear_annotation(msg.path))
        menu.addAction(_('Clear all'), self._clear_annotation)
        menu.addSeparator()
        menu.addAction(_('Open folder'), self._open_folder)
        menu.exec_(event.globalPos())

    def _text_operation(
        self,
        txt,
        insert=True,
        insert_position=None,
        replace_by=''
    ):
        """
        desc:
            Inserts or removes text while preserving the cursor position and
            selection.
        
        arguments:
            text
                The text to insert or remove.
            insert:
                False indicates that the first occurence of `text` should be
                removed.
            insert_position:
                Indicates the position of inserted text. If `None`, then the
                text is inserted at the next line below the current block.
            replace_by:
                Replaces the removed text by a string. Only applicable if
                `insert` is False.
        """
        # Get the cursor and remember the original position and anchor, where
        # the anchor is the start of the selection if any text is selected.
        cursor = self._editor.textCursor()
        # If the code has not changed since the last operation, then we
        # continue the previous edit block. This ensures that all capture
        # operations are treated as a single undo operation.
        code = self._editor.toPlainText()
        if self._code == code:
            cursor.joinPreviousEditBlock()
        else:
            cursor.beginEditBlock()
        restore_selection = cursor.hasSelection()
        anchor = cursor.anchor()
        pos = cursor.position()
        if insert:
            if insert_position is None:
                # If no insert position is specified, we insert after the
                # selected text.
                cursor.setPosition(cursor.selectionEnd())
                if not cursor.atBlockStart() and not cursor.atBlockEnd():
                    cursor.movePosition(cursor.EndOfBlock)
                insert_position = cursor.position()
            else:
                cursor.setPosition(insert_position)
            # Insert a newline if we're not already at the start of a line, so
            # that we don't append to existing lines.
            if not cursor.atBlockStart():
                txt = '\n' + txt
            cursor.insertText(txt)
            # If the original anchor and position were after the inserted text
            # then we need to move them forward.
            if insert_position < anchor:
                anchor += len(txt)
            if insert_position < pos:
                pos += len(txt)
        else:
            # Find the text, select it, and remove it.
            needle_pos = code.find(txt)
            if needle_pos < 0:  # Don't do anything if the text is not found.
                return
            cursor.setPosition(needle_pos)
            cursor.movePosition(
                cursor.Right,
                cursor.KeepAnchor,
                len(txt)
            )
            cursor.removeSelectedText()
            if replace_by:
                cursor.insertText(replace_by)
            # If the original anchor and position were after the inserted text
            # then we need to move them backward.
            if needle_pos < anchor:
                anchor -= len(txt) - len(replace_by)
            if cursor.position() < pos:
                pos -= len(txt) - len(replace_by)
        # Restore the original anchor and position
        cursor.setPosition(anchor)
        if restore_selection:
            cursor.setPosition(pos, cursor.KeepAnchor)
        cursor.endEditBlock()
        self._editor.setTextCursor(cursor)
        self._code = self._editor.toPlainText()

    def settings_widget(self):
        
        from image_annotations import Preferences
        return Preferences(self.main_window)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

KERNEL_NAMES = {
    u'python': u'Python',
    u'python2': u'Python',
    u'python3': u'Python',
    u'ir': u'R',
    u'bash': u'Bash'
}
CHANGE_DIR_CMD = {
    u'ir': 'setwd("{path}")',
    u'bash': 'cd "{path}"'
}
DEFAULT_CHANGE_DIR_CMD = u'%cd "{path}"'
RUN_FILE_CMD = {
    u'ir': 'source("{path}")',
    u'bash': '"{path}"'
}
DEFAULT_RUN_FILE_CMD = u'%run "{path}"'
DEFAULT_RUN_DEBUG_CMD = None
RUN_SYSTEM_CMD = {
    u'ir': 'system("{cmd}")',
    u'bash': '{cmd}'
}
DEFAULT_RUN_SYSTEM_CMD = u'!{cmd}'
TRANSPARENT_KERNELS = [u'python', u'python2', u'python3']
PID_CMD = {
    u'python': u'os.getpid()',
    u'python2': u'os.getpid()',
    u'python3': u'os.getpid()'
}



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import os
import sys
import time
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from libqtopensesame.widgets.base_widget import BaseWidget
from qtpy.QtWidgets import QHBoxLayout
from qtpy.QtGui import QFont
from qtconsole import styles
from jupyter_tabwidget.constants import (
    CHANGE_DIR_CMD,
    DEFAULT_CHANGE_DIR_CMD,
    RUN_FILE_CMD,
    DEFAULT_RUN_FILE_CMD,
    RUN_SYSTEM_CMD,
    DEFAULT_RUN_SYSTEM_CMD,
    TRANSPARENT_KERNELS,
    PID_CMD
)
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'JupyterConsole', category=u'extension')


class JupyterConsole(BaseWidget):

    def __init__(
        self,
        parent=None,
        name=None,
        kernel=None,
        inprocess=False
    ):

        super(JupyterConsole, self).__init__(parent)
        if kernel is None:
            kernel = cfg.jupyter_default_kernel
        self._console_tabwidget = parent
        self.name = name
        # Initialize Jupyter Widget
        if inprocess:
            from qtconsole.inprocess import (
                QtInProcessKernelManager as QtKernelManager
            )
            from jupyter_tabwidget.transparent_jupyter_widget import (
                InprocessJupyterWidget as JupyterWidget
            )
        else:
            from qtconsole.manager import QtKernelManager
            from jupyter_tabwidget.transparent_jupyter_widget import (
                OutprocessJupyterWidget as JupyterWidget
            )
        self._inprocess = inprocess
        self._kernel = kernel
        self._kernel_manager = QtKernelManager(kernel_name=kernel)
        self._kernel_manager.start_kernel()
        self._kernel_client = self._kernel_manager.client()
        self._kernel_client.start_channels()
        self._jupyter_widget = JupyterWidget(self)
        self._jupyter_widget.kernel_manager = self._kernel_manager
        self._jupyter_widget.kernel_client = self._kernel_client
        # Set theme
        self._jupyter_widget._control.setFont(
            QFont(
                cfg.pyqode_font_name,
                cfg.pyqode_font_size
            )
        )
        self._jupyter_widget.style_sheet = styles.sheet_from_template(
            cfg.pyqode_color_scheme,
            u'linux' if styles.dark_style(cfg.pyqode_color_scheme)
            else 'lightbg'
        )
        self._jupyter_widget.syntax_style = cfg.pyqode_color_scheme
        self._jupyter_widget._syntax_style_changed()
        self._jupyter_widget._style_sheet_changed()
        # Add to layout
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(6, 6, 6, 6)
        self._layout.addWidget(self._jupyter_widget)
        self.setLayout(self._layout)

    def capture_stdout(self):

        sys.stdout = self
        sys.stderr = self

    def release_stdout(self):

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    def isatty(self):

        return False
    
    @property
    def pid(self):
        
        if self._inprocess:
            return os.getpid()
        if self._kernel not in PID_CMD:
            return -1
        for i in range(10):
            pid = self._jupyter_widget._silent_execute(PID_CMD[self._kernel])
            if isinstance(pid, int):
                return pid
            time.sleep(0.1)
        oslogger.warning('failed to get pid')
        return -1

    def flush(self):

        pass

    def execute(self, code):

        self._jupyter_widget.execute(code)

    def write(self, msg):

        self._jupyter_widget._control.insertPlainText(
            safe_decode(msg, errors=u'ignore')
        )
        self._jupyter_widget._control.ensureCursorVisible()

    def focus(self):

        self._jupyter_widget._control.setFocus()

    def set_busy(self, busy=True):

        if busy:
            icon = u'os-run'
        else:
            icon = 'utilities-terminal' if self._inprocess else 'os-debug'
        self._set_icon(icon)

    def _set_icon(self, icon):

        self._console_tabwidget.setTabIcon(
            self._console_tabwidget.indexOf(self),
            self.main_window.theme.qicon(icon)
        )

    def restart(self):

        oslogger.debug(u'restarting kernel')
        self._jupyter_widget.request_restart_kernel()
        self._jupyter_widget.reset(clear=True)
        self.extension_manager.fire(
            'workspace_restart',
            name=self.name,
            workspace_func=self.get_workspace_globals
        )
        self.extension_manager.fire(
            'register_subprocess',
            pid=self.pid,
            description='jupyter_console:{}'.format(self.name)
        )

    def interrupt(self):

        oslogger.debug(u'interrupting kernel')
        self._jupyter_widget.request_interrupt_kernel()

    def shutdown(self):

        oslogger.debug(u'shutting down kernel')
        self._jupyter_widget.kernel_client.stop_channels()
        self._jupyter_widget.kernel_manager.shutdown_kernel()

    def show_prompt(self):

        self._jupyter_widget._show_interpreter_prompt()

    def get_workspace_globals(self):

        if self._kernel in TRANSPARENT_KERNELS:
            return self._jupyter_widget.get_workspace_globals()
        return {'not supported': None}

    def list_workspace_globals(self):

        if self._kernel in TRANSPARENT_KERNELS:
            return self._jupyter_widget.list_workspace_globals()
        return []

    def get_workspace_variable(self, name):

        if self._kernel in TRANSPARENT_KERNELS:
            return self._jupyter_widget.get_workspace_variable(name)
        return None
    
    def running(self):
        
        return self._jupyter_widget.running()

    def set_workspace_globals(self, global_dict):

        self._jupyter_widget.set_workspace_globals(global_dict)

    def change_dir(self, path):

        self.execute(
            CHANGE_DIR_CMD.get(
                self._kernel,
                DEFAULT_CHANGE_DIR_CMD
            ).format(path=path.replace(u'\\', u'\\\\'))
        )

    def run_file(self, path):

        self.execute(
            RUN_FILE_CMD.get(
                self._kernel,
                DEFAULT_RUN_FILE_CMD
            ).format(path=path.replace(u'\\', u'\\\\'))
        )
        
    def run_system_command(self, cmd):
        
        tmpl = RUN_SYSTEM_CMD.get(self._kernel, DEFAULT_RUN_SYSTEM_CMD)
        cmd = cmd.replace(u'\\', u'\\\\')
        # If the command has double quotes itself, such as for the R system()
        # command, then we need to escape any quotes in the cmd string.
        if u'"' in tmpl:
            cmd = cmd.replace(u'"', u'\\"')
        self.execute(tmpl.format(cmd=cmd))
        
    def run_debug(self, path, breakpoints):

        # This should be changed into a kernel-agnostic implementation
        if (
            self._kernel not in ('python', 'python2', 'python3') or
            self._inprocess
        ):
            self.extension_manager.fire(
                'notify',
                message=_(u'The {} (inprocess={}) kernel does not support debugging').format(
                    self._kernel,
                    self._inprocess
                )
            )
            return
        code = self.extension_manager.provide(
            'python_debugger_code',
            path=path,
            breakpoints=breakpoints
        )
        if not code:
            oslogger.warning('no Python debugger code received')
            return
        silent_code, nonsilent_code = code
        # RapunzelPDB and debugfile() are silently loaded into the workspace,
        # and then the debugger is callend in a nonsilent way.`
        self._jupyter_widget._kernel_client.execute(silent_code, silent=True)
        self.execute(nonsilent_code)

    def check_syntax(self, code):

        if 'python' not in self._kernel:
            return True
        try:
            compile(code, 'dummy', 'exec')
        except SyntaxError:
            return False
        return True



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import sys
import subprocess
from libqtopensesame.misc.config import cfg
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'IDELauncher', category=u'extension')


class IDELauncher(BaseExtension):

    def activate(self):

        if cfg.ide_executable:
            cmd = cfg.ide_executable
        else:
            cmd = [sys.executable, sys.argv[0], '--mode=ide']
        try:
            subprocess.Popen(cmd)
        except Exception as e:
            self.notify(_(
                u'Failed to launch Code Editor. '
                u'See debug window for error message.'
            ))
            self.console.write(e)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
import importlib
from libqtopensesame.extensions import BaseExtension
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.translate import translation_context
from qtpy.QtWidgets import QDockWidget
from qtpy.QtCore import Qt
from data_viewer_inspectors.inspect_str import inspect_str
import data_viewer_file_handlers as file_handlers
from datadockwidget import DataDockWidget
_ = translation_context(u'DataViewer', category=u'extension')


class DataViewer(BaseExtension):
    
    def event_startup(self):
        
        self._dock_widgets = {}
        self.queue = []

    def event_data_viewer_inspect(self, name, workspace):

        if name in self._dock_widgets:
            return
        self.main_window.set_busy(True)
        value = self.extension_manager.provide(
            'jupyter_workspace_variable',
            name=name
        )
        dw = DataDockWidget(self, name, value, workspace)
        self.main_window.addDockWidget(Qt.RightDockWidgetArea, dw)
        self.main_window.set_busy(False)
        self._dock_widgets[name] = dw
        
    def _provide_ext(self, ext):
        
        """Returns a provider function to handle an extension."""
        
        def inner():

            # Get the [Kernel]FileHandler class based on the kernel of the
            # active workspace
            cls_name = '{}FileHandler'.format(
                self.extension_manager.provide('workspace_kernel').capitalize()
            )
            oslogger.debug('looking for {}'.format(cls_name))
            try:
                cls = getattr(file_handlers, cls_name)
            except AttributeError:
                oslogger.debug('no file handler for kernel')
                return None
            obj = cls(self)
            try:
                # Call the extension provider function, which should return
                # a tuple of a function that actually opens a file, and a
                # descriptive str.
                return getattr(obj, ext)()
            except AttributeError:
                oslogger.debug('no file handler for extension')
                return None
            
        return inner
        
    def supported_provides(self):
        
        """BaseFileHandler has a list of extensions that can be handled, all
        though not all kernel-specific file handlers can handle all extensions.
        For these extensions, a provider function is created.
        """
        
        provides = []
        for ext in file_handlers.BaseFileHandler.supported_extensions:
            # Dynamically set a provider function
            setattr(
                self,
                'provide_open_file_extension_{}'.format(ext),
                self._provide_ext(ext)
            )
            # Create a list of provides
            provides.append('open_file_extension_{}'.format(ext))
        return provides

    def remove_dock_widget(self, name):
        
        if name in self._dock_widgets:
            self.main_window.removeDockWidget(self._dock_widgets[name])
            del self._dock_widgets[name]
            
    def event_workspace_update(self, name, workspace_func):

        self._update(name)
        # The queue contains symbol names that should be shown after the next
        # kernel command has been executed. This allows for loading files from
        # disk with a command, and then subsequently showing them.
        while self.queue:
            symbol = self.queue.pop()
            self.extension_manager.fire(
                'data_viewer_inspect',
                name=symbol,
                workspace=self.extension_manager.provide(
                    'jupyter_workspace_name'
                )
            )

    def event_workspace_restart(self, name, workspace_func):

        self._update(name)

    def event_workspace_switch(self, name, workspace_func):

        self._update(name)

    def event_workspace_new(self, name, workspace_func):

        self._update(name)
               
    def _update(self, name):

        # Only get the data for the current workspace
        dock_widgets = {
            name: dw
            for name, dw in self._dock_widgets.items()
            if dw.workspace != name
        }
        if not dock_widgets:
            return
        self.main_window.set_busy(True)
        workspace_list = None
        # Loop through all dockwidgets and get the current value of the
        # variable
        for name, dw in dock_widgets.items():
            value = self.extension_manager.provide(
                'jupyter_workspace_variable',
                name=name
            )
            # If the value is None, this can either mean that it is really None
            # or that the variable no longer exists. To check this, we get a
            # list of all global variables. However, for performance, we do
            # this only once, and only if it's necessary.
            if value is None:
                if workspace_list is None:
                    workspace_list = self.extension_manager.provide(
                        'jupyter_list_workspace_globals'
                    )
                # If the variable has been deleter, close the corresponding
                # dockwidget
                if name not in workspace_list:
                    dw.close()
                    continue
            dw.refresh(value)
        self.main_window.set_busy(False)



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""

from libopensesame.py3compat import *
from jupyter_notebook_cell_parsers.parse_python import FENCED_BLOCK_RE


def parse_markdown(code=u'', cell_types=None):
    
    cells = []
    offset = 0
    while True:
        m = FENCED_BLOCK_RE.search(code)
        if not m:
            break
        chunk = code[m.start():m.end()]
        start = m.start() + offset + chunk.find(u'\n') + 1
        end = m.start() + offset + chunk.rfind(u'\n')
        cells.append({
            'cell_type': 'code',
            'source': m.group('code'),
            'start': start,
            'end': end
        })
        offset += m.end()
        code = code[m.end():]
    return cells
